-------------------------------------
--  lxi-tools                      --
--    https://lxi-tools.github.io  --
-------------------------------------

-- Automation test



-- tc "Test case 1 description"

print("This is test case 1")



-- tc "Test case 2 description"

print("This is test case 2")

fail()



-- tc "Test case 3 description"

print("This is test case 3")

fail_stop()



-- tc "Test case 4 description"

print("This is test case 4")



-- tc save test-results.log
